@@ -1,3 +1,5 @@
+import copy
+
 from ipywidgets import (
     Widget, DOMWidget, Box, Color, CallbackDispatcher, widget_serialization,
     interactive
@@ -8,6 +10,8 @@ from traitlets import (
     link, observe, default, validate, TraitError
 )
 
+from branca.colormap import linear, ColorMap
+
 from traittypes import Dataset
 
 from .xarray_ds import ds_x_to_json
@@ -73,10 +77,55 @@ class Layer(Widget, InteractMixin):
 
     options = List(trait=Unicode).tag(sync=True)
 
+    def __init__(self, **kwargs):
+        super(Layer, self).__init__(**kwargs)
+        self.on_msg(self._handle_mouse_events)
+
     @default('options')
     def _default_options(self):
         return [name for name in self.traits(o=True)]
 
+    # Event handling
+    _click_callbacks = Instance(CallbackDispatcher, ())
+    _dblclick_callbacks = Instance(CallbackDispatcher, ())
+    _mousedown_callbacks = Instance(CallbackDispatcher, ())
+    _mouseup_callbacks = Instance(CallbackDispatcher, ())
+    _mouseover_callbacks = Instance(CallbackDispatcher, ())
+    _mouseout_callbacks = Instance(CallbackDispatcher, ())
+
+    def _handle_mouse_events(self, _, content, buffers):
+        event_type = content.get('type', '')
+        if event_type == 'click':
+            self._click_callbacks(**content)
+        if event_type == 'dblclick':
+            self._dblclick_callbacks(**content)
+        if event_type == 'mousedown':
+            self._mousedown_callbacks(**content)
+        if event_type == 'mouseup':
+            self._mouseup_callbacks(**content)
+        if event_type == 'mouseover':
+            self._mouseover_callbacks(**content)
+        if event_type == 'mouseout':
+            self._mouseout_callbacks(**content)
+
+    def on_click(self, callback, remove=False):
+        self._click_callbacks.register_callback(callback, remove=remove)
+
+    def on_dblclick(self, callback, remove=False):
+        self._dblclick_callbacks.register_callback(callback, remove=remove)
+
+    def on_mousedown(self, callback, remove=False):
+        self._mousedown_callbacks.register_callback(callback, remove=remove)
+
+    def on_mouseup(self, callback, remove=False):
+        self._mouseup_callbacks.register_callback(callback, remove=remove)
+
+    def on_mouseover(self, callback, remove=False):
+        self._mouseover_callbacks.register_callback(callback, remove=remove)
+
+    def on_mouseout(self, callback, remove=False):
+        self._mouseout_callbacks.register_callback(callback, remove=remove)
+
 
 class UILayer(Layer):
     _view_name = Unicode('LeafletUILayerView').tag(sync=True)
@@ -434,11 +483,50 @@ class GeoJSON(FeatureGroup):
         self._hover_callbacks.register_callback(callback, remove=remove)
 
 
-class MultiPolygon(FeatureGroup):
-    _view_name = Unicode('LeafletMultiPolygonView').tag(sync=True)
-    _model_name = Unicode('LeafletMultiPolygonModel').tag(sync=True)
+class Choropleth(GeoJSON):
 
-    locations = List().tag(sync=True)
+    geo_data = Dict()
+    choro_data = Dict()
+    value_min = Float(None, allow_none=True)
+    value_max = Float(None, allow_none=True)
+    colormap = Instance(ColorMap)
+    border_color = Color('black')
+
+    @observe('choro_data')
+    def _update_bounds(self, change):
+        self.value_min = min(self.choro_data.items(), key=lambda x: x[1])[1]
+        self.value_max = max(self.choro_data.items(), key=lambda x: x[1])[1]
+
+    @observe('value_min', 'value_max', 'geo_data', 'choro_data', 'colormap', 'border_color')
+    def _update_data(self, change):
+        self.data = self._get_data()
+
+    @default('colormap')
+    def _default_colormap(self):
+        return linear.OrRd_06
+
+    def _get_data(self):
+        if not self.geo_data:
+            return {}
+
+        if self.value_min is None:
+            self.value_min = min(self.choro_data.items(), key=lambda x: x[1])[1]
+        if self.value_max is None:
+            self.value_max = max(self.choro_data.items(), key=lambda x: x[1])[1]
+
+        colormap = self.colormap.scale(self.value_min, self.value_max)
+        color_dict = {key: colormap(self.choro_data[key]) for key in self.choro_data.keys()}
+
+        data = copy.deepcopy(self.geo_data)
+        for feature in data['features']:
+            feature['properties']['style'] = dict(fillColor=color_dict[feature['id']],
+                                                  color=self.border_color,
+                                                  weight=0.9)
+        return data
+
+    def __init__(self, **kwargs):
+        super(Choropleth, self).__init__(**kwargs)
+        self.data = self._get_data()
 
 
 class ControlException(TraitError):
